from .obj_reader import ObjToTriangles
from .calculations import normalize as VecNorm
from .calculations import point_distance as VecDistance
from .calculations import vector_angle as VecAngle
from .calculations import vector_inverse as VecInv
from .calculations import position_between_xz as PosBetweenXZ
from .calculations import sort_nearest_points_from_on_plane_y as SortPointsFromPlaneY
from .converters import outdoor_traced_result_to_line as OutdoorResultToLines



from pyglet.window import Window, key
from pyglet.gl import glClearColor, glBegin, glEnd, GL_LINE_STRIP,\
    glVertex3f, glColor4f, glLineWidth, glMatrixMode, GL_PROJECTION,\
    glLoadIdentity, gluPerspective, GL_MODELVIEW, glTranslatef, glRotatef,\
    glEnable, GL_DEPTH_TEST, glDrawArrays, GL_TRIANGLES, glPushMatrix

import logging
from pywavefront import visualization, Wavefront, configure_logging

configure_logging(
    logging.ERROR,
    formatter=logging.Formatter('%(name)s-%(levelname)s: %(message)s')
)

class VisualizerWindow(Window):
    camera_position = None
    camera_rotation = None
    window_size = None
    resizable = False
    scene = []
    objects = []
    alive = False
    pressed_keys = {}
    zoom = 0
    clear_color = (0,0,0,1)
    line_sets = []
    scenes = []
    def __init__(self, title = "Radio Gyms", window_size = (600, 400),
                 camera_position=[0 , 0, -255],
                 camera_rotation=[33, -230, 0],
                 resizable=True,
                 background_color=[0,0,0,1],
                 zoom = 66):

        self.window_size = window_size
        self.camera_position = camera_position
        self.camera_rotation = camera_rotation
        self.resizable = resizable
        self.alive = True
        self.pressed_keys = {}
        self.zoom = zoom
        self.clear_color = background_color
        self.line_sets = []
        self.scenes = []

        super().__init__(width=window_size[0],
                         height=window_size[1],
                         resizable = resizable,
                         caption = title,
                         style=Window.WINDOW_STYLE_DIALOG)
        glClearColor(0.9, 0.9, 0.875, 1)

    def load_icon(self):
        pass

    def on_key_press(self, symbol, modifiers):
        if symbol == key.ESCAPE:
            self.alive = False
        self.pressed_keys[symbol] = True

    def on_key_release(self, symbol, modifiers):
        try:
            del self.pressed_keys[symbol]
        except:
            pass

    def load_obj_to_scene(self, file_path):
        scene = Wavefront(file_path)
        self.scenes.append(scene)

    @staticmethod
    def draw_line(line_info):
        points = line_info['points']
        color = line_info['color']
        glBegin(GL_LINE_STRIP)
        glColor4f(*color)
        for point in points:
            glVertex3f(*point)
        glEnd()
        glLineWidth(2)

    def key_updates(self):
        if key.W in self.pressed_keys:
            self.camera_position[2] += 1
        if key.S in self.pressed_keys:
            self.camera_position[2] -= 1
        if key.A in self.pressed_keys:
            self.camera_position[0] += 1
        if key.D in self.pressed_keys:
            self.camera_position[0] -= 1
        if key.LSHIFT in self.pressed_keys:
            self.camera_position[1] += 1
        if key.LCTRL in self.pressed_keys:
            self.camera_position[1] -= 1
        if key.Z in self.pressed_keys:
            self.camera_rotation[0] += 1
        if key.X in self.pressed_keys:
            self.camera_rotation[0] -= 1
        if key.Q in self.pressed_keys:
            self.camera_rotation[1] += 1
        if key.E in self.pressed_keys:
            self.camera_rotation[1] -= 1

    def on_close(self):
        self.alive = False

    def on_draw(self):
        self.render()

    def render(self):
        self.clear()
        self.key_updates()
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(self.zoom, self.window_size[0] /
                       float(self.window_size[1]), .1, 1000)
        glMatrixMode(GL_MODELVIEW)
        glDrawArrays(GL_TRIANGLES, 0, 3)
        glLoadIdentity()
        glTranslatef(*self.camera_position)
        glRotatef(self.camera_rotation[0], 1.0, 0.0, 0.0)
        glRotatef(self.camera_rotation[1], 0.0, 1.0, 0.0)
        glRotatef(self.camera_rotation[2], 0.0, 0.0, 1.0)
        glEnable(GL_DEPTH_TEST)
        for scene in self.scenes:
            visualization.draw(scene)
        for lineInfo in self.line_sets:
            self.draw_line(lineInfo)
        self.flip()

    def run(self):
        while self.alive:
            event = self.dispatch_events()
            self.render()


import unittest
import radio_gyms


class TestVersion(unittest.TestCase):

    def test_version(self):
        """Check Version from radio_gyms"""
        self.assertEqual(radio_gyms.__version__, "0.1.1")


if __name__ == '__main__':
    unittest.main()


from typing import List, Tuple

class TheoreticalOutdoorModel:
    """
    Theoretical Outdoor model consist of free-space loss, reflection loss, and knife-edge diffraction.
    This theoretical outdoor model can be used for ultra-high frequency up to 100 GHz.

    The ray tracing result can be calculated for loss and signal propagation delay
    """
    result = None

    def __init__(self, result):
        self.result = result

    def calculatePropagationLoss(self) -> float:
        pass

    def calculatePropagationLoss(self) -> List[Tuple[float, float]]:
        pass


import numpy as np
from typing import List
from numpy.typing import NDArray
from .constants import EPSILON


def normalize(vector: NDArray) -> NDArray:
    norm = np.linalg.norm(vector)
    if norm != 0:
        vector = vector / norm
    return vector


def point_distance(point_a: NDArray, point_b: NDArray) -> float:
    return np.sqrt(np.sum((point_a - point_b) ** 2))


def vector_angle(a_dir: NDArray, b_dir: NDArray) -> float:
    a_norm = normalize(a_dir)
    b_norm = normalize(b_dir)
    dot_product = np.dot(a_norm, b_norm)
    return np.arccos(dot_product)


def position_between_xz(min_x, max_x, min_z, max_z, pos) -> bool:
    return min_x <= pos[0] <= max_x and min_z <= pos[2] <= max_z


def vector_inverse(vector: NDArray) -> NDArray:
    inv_vec = np.copy(vector).astype('float32')
    for i in range(len(vector)):
        if inv_vec[i] == 0:
            inv_vec[i] = EPSILON
    return 1.0 / inv_vec


def plane_y_distance(pos_a: NDArray, pos_b: NDArray):
    a_on_plane, b_on_plane = np.copy(pos_a), np.copy(pos_b)
    a_on_plane[1], b_on_plane[1] = 0.0, 0.0
    return point_distance(a_on_plane, b_on_plane)


def sort_nearest_points_from_on_plane_y(ref_pos: NDArray, points: List[NDArray]) -> List[NDArray]:
    sorted_points = sorted(points, key=lambda point: plane_y_distance(point, ref_pos))
    return sorted_points



from typing import List


def outdoor_traced_result_to_line(result, tx_pos, rx_pos) -> List[List]:
    lines = []
    if result['direct']:
        line = {'points': [tx_pos, rx_pos], 'color': [0, 0.8, 0, 1]}
        lines.append(line)
    else:
        if len(result['roof_edges']) > 0:
            line = {'points': [tx_pos] + result['roof_edges'] + [rx_pos], 'color': [0.2, 0.2, 0.2, 1]}
            lines.append(line)

    if 'reflections' in result:
        if 'single' in result['reflections']:
            for reflect_point in result['reflections']['single']:
                line = {'points': [tx_pos, reflect_point, rx_pos], 'color': [0, 0, 0.7, 1.0]}
                # lines.append(line)
        if 'double' in result['reflections']:
            for reflect_points in result['reflections']['double']:
                line = {'points': [tx_pos] + reflect_points + [rx_pos], 'color': [0.3, .3, 0.75, 1.0]}
                lines.append(line)
    return lines



import numpy as np
from typing import Sequence
from ..engines.ray_tracer.triangle import Triangle


def ObjToTriangles(file_path: str) -> Sequence[Triangle]:
    faces: Sequence[int] = []
    vertices: Sequence[float] = []
    triangles: Sequence[Triangle] = []

    with open(file_path, "r") as file:
        for line in file:
            components = line.strip(" \n").split(" ")
            if components[0] == "f":
                indices = list(
                    map(lambda c: int(c.split('/')[0]) - 1, components[1:]))
                for i in range(0, len(indices) - 2):
                    faces.append(indices[i: i + 3])
            elif components[0] == "v":
                vertex = list(map(lambda c: float(c), components[1:]))
                vertices.append(vertex)

    for face in faces:
        a = np.array(vertices[face[0]])
        b = np.array(vertices[face[1]])
        c = np.array(vertices[face[2]])
        triangle = Triangle(a, b, c)
        triangles.append(triangle)
    return triangles



import os
import numpy as np
from unittest import TestCase
from radio_gyms.visualizers import Window
from radio_gyms.engines.ray_tracer.tracer import Tracer
from radio_gyms.utils.converters import outdoor_traced_result_to_line as OutdoorResultToLines

POZNAN_OBJ_PATH = os.path.join(os.getcwd(), "assets", "models", "poznan.obj")

class TestVirtualizer(TestCase):

    def test_result_display(self):
        window = Window()
        tracer = Tracer(POZNAN_OBJ_PATH)
        tx_pos = np.array([0, 5, 0])
        lines = []
        for i in range(1):
            while True:
                rx_pos = (np.random.rand(3)*2-1)*100
                rx_pos[1] = 1.2
                if tracer.is_outdoor(rx_pos):
                    break
            result = tracer.trace_outdoor(tx_pos, rx_pos)
            lines = lines + OutdoorResultToLines(result, tx_pos, rx_pos)
        window.line_sets = lines
        window.load_obj_to_scene(POZNAN_OBJ_PATH)
        window.run()


from typing import Tuple, List, Dict
from numpy.typing import NDArray

import numpy as np

from ...utils import ObjToTriangles, VecNorm, VecDistance, VecAngle, PosBetweenXZ, SortPointsFromPlaneY
from ...utils.constants import EPSILON, MIN_ROOF_EDGE_DISTANCE, ROOF_MIN_ANGLE, ROOF_MAX_SCAN, MAX_FLT
from ...utils import VecNorm
from .bvh import BVH


class Tracer:

    def __init__(self, object_file_path):
        """
        Initialize Map for Ray Tracer
        :param object_file_path:
        """
        triangles = ObjToTriangles(object_file_path)
        self.map = BVH(triangles)

    def trace_outdoor(self, tx_pos: List[float], rx_pos: List[float]):
        """
        Trace the possible ray paths from tx_pos to rx_pos in outdoor scenario (open sky)
        :param tx_pos: Transmitting Position
        :param rx_pos: Receiving Position
        :return: Outdoor Tracing Result
        """
        tx_pos = np.array(tx_pos)
        rx_pos = np.array(rx_pos)
        result = {
            "direct": True,
            "reflections": [],
            "roofEdges": [],
            'tx_pos': tx_pos,
            'rx_pos': rx_pos
        }

        if self.direct_path(tx_pos, rx_pos):
            result["direct"] = True
        else:
            result["direct"] = False
            sorted_edges = SortPointsFromPlaneY(tx_pos, self.trace_roof_edges(tx_pos, rx_pos))
            result["roof_edges"] = sorted_edges

        result['reflections'] = self.trace_reflections(tx_pos, rx_pos)
        return result

    @staticmethod
    def make_ray(tx_pos: NDArray, rx_pos: NDArray) -> Tuple[NDArray, NDArray]:
        ray_org: NDArray = tx_pos
        ray_dir: NDArray = VecNorm(rx_pos - tx_pos)
        ray = (ray_org, ray_dir)
        return ray

    def direct_path(self, tx_pos: NDArray, rx_pos: NDArray):
        """
        Check Direct Path
        :param tx_pos: Transmitting Position
        :param rx_pos: Receiving Position
        :return: true if tx_pos and rx_pos are in line-of-sight.
        """
        ray = Tracer.make_ray(tx_pos, rx_pos)
        point_distance = VecDistance(tx_pos, rx_pos)
        nearest_hit = self.map.is_intersect(ray)
        if nearest_hit < 0:
            return True
        return nearest_hit > point_distance

    @staticmethod
    def get_mirror_point(pos: NDArray, triangle: 'Triangle') -> NDArray:
        normal = triangle.normal
        b = np.dot(normal, triangle.pointB)
        c = np.dot(pos, normal)
        d = np.dot(normal, normal)
        if d == 0:
            d = EPSILON
        t = (b - c) / d
        return pos + normal * 2 * t

    def trace_reflections(self, tx_pos: NDArray, rx_pos: NDArray) -> Dict:
        """
        Trace Reflection Points
        :param tx_pos: Transmitting Position
        :param rx_pos: Receiving Position
        :return: reflection points
        """
        reflections = {
            'single': self.trace_single_reflect(tx_pos, rx_pos),
            'double': self.trace_double_reflect(tx_pos, rx_pos)
        }
        return reflections

    def trace_single_reflect(self, tx_pos, rx_pos) -> List[NDArray]:
        single_reflections = []
        for triangle in self.map.root.triangles:
            mirror_point = Tracer.get_mirror_point(tx_pos, triangle)
            dir_mirror_to_rx = VecNorm(rx_pos - mirror_point)
            ray = (mirror_point, dir_mirror_to_rx)
            mirror_to_rx_dist = triangle.is_intersect(ray)
            if mirror_to_rx_dist < 0:
                continue
            point_on_triangle = mirror_point + dir_mirror_to_rx * (mirror_to_rx_dist + EPSILON)
            if self.direct_path(tx_pos, point_on_triangle) and \
                    self.direct_path(rx_pos, point_on_triangle):
                single_reflections.append(point_on_triangle)
        return single_reflections

    def trace_double_reflect(self, tx_pos, rx_pos):
        double_reflections = []
        tx_mirror_points = []
        rx_mirror_points = []
        triangle_n = len(self.map.root.triangles)
        for triangle in self.map.root.triangles:
            tx_mirror_point = Tracer.get_mirror_point(tx_pos, triangle)
            rx_mirror_point = Tracer.get_mirror_point(rx_pos, triangle)
            tx_mirror_points.append(tx_mirror_point)
            rx_mirror_points.append(rx_mirror_point)

        for tx_mirror_i in range(triangle_n):
            tx_mirror_point = tx_mirror_points[tx_mirror_i]
            tx_triangle = self.map.root.triangles[tx_mirror_i]
            for rx_mirror_i in range(tx_mirror_i + 1, triangle_n):
                rx_mirror_point = rx_mirror_points[rx_mirror_i]
                rx_triangle = self.map.root.triangles[rx_mirror_i]

                tx_mirror_to_rx_mirror_dir = VecNorm(rx_mirror_point - tx_mirror_point)
                rx_mirror_to_tx_mirror_dir = -tx_mirror_to_rx_mirror_dir

                tx_mirror_ray = (tx_mirror_point, tx_mirror_to_rx_mirror_dir)
                rx_mirror_ray = (rx_mirror_point, rx_mirror_to_tx_mirror_dir)

                tx_mirror_to_rx_mirror_dist = tx_triangle.is_intersect(tx_mirror_ray)
                rx_mirror_to_tx_mirror_dist = rx_triangle.is_intersect(rx_mirror_ray)
                if tx_mirror_to_rx_mirror_dist < 0 or rx_mirror_to_tx_mirror_dist < 0:
                    continue
                tx_point_on_triangle = tx_mirror_point + tx_mirror_to_rx_mirror_dir * (
                        tx_mirror_to_rx_mirror_dist + EPSILON)
                rx_point_on_triangle = rx_mirror_point + rx_mirror_to_tx_mirror_dir * (
                        rx_mirror_to_tx_mirror_dist + EPSILON)

                if self.direct_path(tx_pos, tx_point_on_triangle) and \
                        self.direct_path(tx_point_on_triangle, rx_point_on_triangle) and \
                        self.direct_path(rx_point_on_triangle, rx_pos):
                    double_reflections.append([tx_point_on_triangle, rx_point_on_triangle])
        return double_reflections

    def trace_roof_edges(self, tx_pos, rx_pos) -> List:
        """
        Trace Knife Edges
        :param tx_pos: Transmitting Position
        :param rx_pos: Receiving Position
        :return: Knife Edges
        """
        edges = []
        left_pos = tx_pos
        right_pos = rx_pos

        current_scan = 0

        while not self.direct_path(left_pos, right_pos):
            if current_scan > ROOF_MAX_SCAN:
                return edges
            edge_left = self.find_edge(left_pos, right_pos)
            if edge_left is None:
                return []
            edge_right = self.find_edge(right_pos, left_pos)
            if edge_right is None:
                return []

            if self.direct_path(edge_left, edge_right):
                if VecDistance(edge_left, edge_right) < MIN_ROOF_EDGE_DISTANCE:
                    avg_edge = (edge_left + edge_right) / 2
                    edges.append(avg_edge)
                    return edges
                edges.append(edge_left)
                edges.append(edge_right)
                return edges

            edges.append(edge_left)
            edges.append(edge_right)
            left_pos = edge_left
            right_pos = edge_right
        return []

    def find_edge(self, left_pos, right_pos):
        min_x = min(left_pos[0], right_pos[0])
        max_x = max(left_pos[0], right_pos[0])
        min_z = min(left_pos[2], right_pos[2])
        max_z = max(left_pos[2], right_pos[2])

        top_direction = np.array([0, 1, 0])
        upper_ray = (left_pos, top_direction)
        if self.map.is_intersect(upper_ray) > 0:
            return None
        lower_ray = (left_pos, VecNorm(right_pos - left_pos))

        current_scan = 0
        while current_scan < ROOF_MAX_SCAN and \
                VecAngle(upper_ray[1], lower_ray[1]) > ROOF_MIN_ANGLE:
            current_scan += 1

            new_dir = VecNorm((upper_ray[1] + lower_ray[1]) / 2)
            check_ray = (left_pos, new_dir)

            hit_nearest = self.map.is_intersect(check_ray)
            if hit_nearest > 0 and PosBetweenXZ(min_x, max_x, min_z, max_z, left_pos + new_dir * hit_nearest):
                lower_ray = check_ray
            else:
                upper_ray = check_ray

        distance = self.map.is_intersect(lower_ray)
        if distance < 0:
            return None

        left_pos_on_plane = np.array([left_pos[0], 0, left_pos[2]])
        right_pos_on_plane = np.array([right_pos[0], 0, right_pos[2]])
        plane_dir = VecNorm(right_pos_on_plane - left_pos_on_plane)

        theta = VecAngle(plane_dir, lower_ray[1])
        x_angle = VecAngle(lower_ray[1], upper_ray[1])

        height = distance * np.cos(theta) * np.tan(theta + x_angle)
        width = distance * np.cos(theta)
        edge_distance = np.sqrt(height ** 2 + width ** 2)
        edge_pos = upper_ray[0] + upper_ray[1] * edge_distance
        return edge_pos

    def is_outdoor(self, pos):
        sky_pos = np.copy(pos)
        sky_pos[1] = 1000
        return self.direct_path(pos, sky_pos)



from typing import Sequence, Tuple, List
from numpy.typing import NDArray

from .triangle import Triangle
from ...utils.constants import MAX_FLT
from .box import Box


class BVH:
    root: Box = None

    def __init__(self, triangles: Sequence[Triangle]):
        self.root = Box(triangles)
        BVH.make_children(self.root, 0)

    def is_intersect(self, ray: Tuple[NDArray, NDArray]) -> float:
        hot_boxes = [self.root]
        hit_once = False
        nearest_hit = MAX_FLT

        while len(hot_boxes) != 0:
            current_box = hot_boxes.pop(0)

            if current_box.alone:
                current_hit = current_box.triangles[0].is_intersect(ray)
                if current_hit > 0:
                    nearest_hit = min(nearest_hit, current_hit)
                    hit_once = True
                continue

            if not current_box.is_intersect(ray):
                continue
            if current_box.child_left is not None:
                hot_boxes.append(current_box.child_left)
            if current_box.child_right is not None:
                hot_boxes.append(current_box.child_right)

        if hit_once:
            return nearest_hit
        return -1

    @staticmethod
    def make_children(parent: Box, level: int):
        k: int = level % 3
        ordered_triangles = []
        if k == 0:
            ordered_triangles = sorted(parent.triangles, key=lambda tri: tri.max_x, reverse=False)
        elif k == 1:
            ordered_triangles = sorted(parent.triangles, key=lambda tri: tri.max_y, reverse=False)
        elif k == 2:
            ordered_triangles = sorted(parent.triangles, key=lambda tri: tri.max_z, reverse=False)

        middle_index = int(len(ordered_triangles) / 2)
        left_triangles = ordered_triangles[:middle_index]
        right_triangles = ordered_triangles[middle_index:]

        n_left = len(left_triangles)
        n_right = len(right_triangles)

        level += 1
        if n_left != 0:
            parent.child_left = Box(left_triangles)
            if n_left > 1:
                BVH.make_children(parent.child_left, level)
        if n_right != 0:
            parent.child_right = Box(right_triangles)
            if n_right > 1:
                BVH.make_children(parent.child_right, level)




import numpy as np

from typing import Tuple
from numpy.typing import NDArray

from ...utils.calculations import normalize as VecNorm
from ...utils.constants import EPSILON


class Triangle:
    pointA: NDArray = None
    pointB: NDArray = None
    pointC: NDArray = None
    normal: NDArray = None
    edgeA: NDArray = None
    edgeB: NDArray = None
    min_x: float = None
    max_x: float = None
    min_y: float = None
    max_y: float = None
    min_z: float = None
    max_z: float = None

    def __init__(self, a: NDArray, b: NDArray, c: NDArray):
        self.pointA = a
        self.pointB = b
        self.pointC = c
        self.edgeA = self.pointB - self.pointA
        self.edgeB = self.pointC - self.pointA
        self.normal = VecNorm(np.cross(self.edgeA, self.edgeB))
        self.min_x = min([a[0], b[0], c[0]])
        self.min_y = min([a[1], b[1], c[1]])
        self.min_z = min([a[2], b[2], c[2]])
        self.max_x = max([a[0], b[0], c[0]])
        self.max_y = max([a[1], b[1], c[1]])
        self.max_z = max([a[2], b[2], c[2]])

    def is_intersect(self, ray: Tuple[NDArray, NDArray]) -> float:
        ray_pos: NDArray = ray[0]
        ray_dir: NDArray = ray[1]

        h = np.cross(ray_dir, self.edgeB)
        a = np.dot(self.edgeA, h)
        if -EPSILON < a < EPSILON:
            return -1
        f = 1.0 / float(a)
        s = ray_pos - self.pointA
        u = f * np.dot(s, h)
        if u < 0 or u > 1:
            return -1

        q = np.cross(s, self.edgeA)
        v = f * np.dot(ray_dir, q)
        if v < 0 or (u + v) > 1:
            return -1

        distance = f * np.dot(self.edgeB, q)
        if distance > EPSILON:
            return distance
        return -1
